#include "vector/vec.h"
#include "math/math_utils.h"
#include <math.h>

vec2i_t vec_add_vec2i_t(vec2i_t v1, vec2i_t v2) {
  return (vec2i_t){.x = v1.x + v2.x, .y = v1.y + v2.y};
}

vec2i_t vec_sub_vec2i_t(vec2i_t v1, vec2i_t v2) {
  return (vec2i_t){.x = v1.x - v2.x, .y = v1.y - v2.y};
}

vec2i_t vec_mul_vec2i_t(vec2i_t v1, vec2i_t v2) {
  return (vec2i_t){.x = v1.x * v2.x, .y = v1.y * v2.y};
}

vec2i_t vec_mul_num_vec2i_t(vec2i_t v1, i32 num) {
  return (vec2i_t){.x = v1.x * num, .y = v1.y * num};
}

vec2i_t vec_div_num_vec2i_t(vec2i_t v1, i32 num) {
  return (vec2i_t){.x = v1.x / num, .y = v1.y / num};
}

i32 vec_dot_vec2i_t(vec2i_t v1, vec2i_t v2) {
  return v1.x * v2.x + v1.y * v2.y;
}

i32 vec_magnitude_vec2i_t(vec2i_t v) {
  i32 dot_product = vec_dot_vec2i_t(v, v);
  return (i32)sqrtf((f32)dot_product);
}

vec2i_t vec_unit_vec2i_t(vec2i_t v) {
  i32 magnitude = vec_magnitude_vec2i_t(v);
  return (vec2i_t){v.x / magnitude, v.y / magnitude};
}

vec2f_t vec_add_vec2f_t(vec2f_t v1, vec2f_t v2) {
  return (vec2f_t){.x = v1.x + v2.x, .y = v1.y + v2.y};
}

vec2f_t vec_sub_vec2f_t(vec2f_t v1, vec2f_t v2) {
  return (vec2f_t){.x = v1.x - v2.x, .y = v1.y - v2.y};
}

vec2f_t vec_mul_vec2f_t(vec2f_t v1, vec2f_t v2) {
  return (vec2f_t){.x = v1.x * v2.x, .y = v1.y * v2.y};
}

vec2f_t vec_mul_num_vec2f_t(vec2f_t v1, f32 num) {
  return (vec2f_t){.x = v1.x * num, .y = v1.y * num};
}

vec2f_t vec_div_num_vec2f_t(vec2f_t v1, f32 num) {
  return (vec2f_t){.x = v1.x / num, .y = v1.y / num};
}

f32 vec_dot_vec2f_t(vec2f_t v1, vec2f_t v2) {
  return v1.x * v2.x + v1.y * v2.y;
}

f32 vec_magnitude_vec2f_t(vec2f_t v) {
  f32 dot_product = vec_dot_vec2f_t(v, v);
  return sqrtf(dot_product);
}

vec2f_t vec_unit_vec2f_t(vec2f_t v) {
  f32 magnitude = vec_magnitude_vec2f_t(v);
  return (vec2f_t){v.x / magnitude, v.y / magnitude};
}

vec3i_t vec_add_vec3i_t(vec3i_t v1, vec3i_t v2) {
  return (vec3i_t){.x = v1.x + v2.x, .y = v1.y + v2.y, .z = v1.z + v2.z};
}

vec3i_t vec_sub_vec3i_t(vec3i_t v1, vec3i_t v2) {
  return (vec3i_t){.x = v1.x - v2.x, .y = v1.y - v2.y, .z = v1.z - v2.z};
}

vec3i_t vec_mul_vec3i_t(vec3i_t v1, vec3i_t v2) {
  return (vec3i_t){.x = v1.x * v2.x, .y = v1.y * v2.y, .z = v1.z * v2.z};
}

vec3i_t vec_mul_num_vec3i_t(vec3i_t v1, i32 num) {
  return (vec3i_t){.x = v1.x * num, .y = v1.y * num, .z = v1.z * num};
}

vec3i_t vec_div_num_vec3i_t(vec3i_t v1, i32 num) {
  return (vec3i_t){.x = v1.x / num, .y = v1.y / num, .z = v1.z / num};
}

i32 vec_dot_vec3i_t(vec3i_t v1, vec3i_t v2) {
  return v1.x * v2.x + v1.y * v2.y + v1.z * v2.z;
}

i32 vec_magnitude_vec3i_t(vec3i_t v) {
  i32 dot_product = vec_dot_vec3i_t(v, v);
  return (i32)sqrtf((f32)dot_product);
}

vec3i_t vec_unit_vec3i_t(vec3i_t v) {
  i32 magnitude = vec_magnitude_vec3i_t(v);
  return (vec3i_t){v.x / magnitude, v.y / magnitude, v.z / magnitude};
}

vec3f_t vec_add_vec3f_t(vec3f_t v1, vec3f_t v2) {
  return (vec3f_t){.x = v1.x + v2.x, .y = v1.y + v2.y, .z = v1.z + v2.z};
}

vec3f_t vec_sub_vec3f_t(vec3f_t v1, vec3f_t v2) {
  return (vec3f_t){.x = v1.x - v2.x, .y = v1.y - v2.y, .z = v1.z - v2.z};
}

vec3f_t vec_mul_num_vec3f_t(vec3f_t v1, f32 num) {
  return (vec3f_t){.x = v1.x * num, .y = v1.y * num, .z = v1.z * num};
}

vec3f_t vec_div_num_vec3f_t(vec3f_t v1, f32 num) {
  return (vec3f_t){.x = v1.x / num, .y = v1.y / num, .z = v1.z / num};
}

vec3f_t vec_mul_vec3f_t(vec3f_t v1, vec3f_t v2) {
  return (vec3f_t){.x = v1.x * v2.x, .y = v1.y * v2.y, .z = v1.z * v2.z};
}

f32 vec_dot_vec3f_t(vec3f_t v1, vec3f_t v2) {
  return v1.x * v2.x + v1.y * v2.y + v1.z * v2.z;
}

f32 vec_magnitude_vec3f_t(vec3f_t v) {
  f32 dot_product = vec_dot_vec3f_t(v, v);
  return sqrtf(dot_product);
}

vec3f_t vec_unit_vec3f_t(vec3f_t v) {
  f32 magnitude = vec_magnitude_vec3f_t(v);
  return (vec3f_t){v.x / magnitude, v.y / magnitude, v.z / magnitude};
}

vec3f_t vec_rotate_vec3f_t(vec3f_t original, vec3f_t rotation) {
  mat3x3f_t matrix = get_rotation_mat3x3f(rotation);

  // clang-format off
	return (vec3f_t){
		.x = matrix.row0.x * original.x + matrix.row0.y * original.y + matrix.row0.z * original.z,
		.y = matrix.row1.x * original.x + matrix.row1.y * original.y + matrix.row1.z * original.z,
		.z = matrix.row2.x * original.x + matrix.row2.y * original.y + matrix.row2.z * original.z,
	};
  // clang-format on
}

vec4i_t vec_add_vec4i_t(vec4i_t v1, vec4i_t v2) {
  return (vec4i_t){
      .x = v1.x + v2.x, .y = v1.y + v2.y, .z = v1.z + v2.z, .w = v1.w + v2.w};
}

vec4i_t vec_sub_vec4i_t(vec4i_t v1, vec4i_t v2) {
  return (vec4i_t){
      .x = v1.x - v2.x, .y = v1.y - v2.y, .z = v1.z - v2.z, .w = v1.w - v2.w};
}

vec4i_t vec_mul_vec4i_t(vec4i_t v1, vec4i_t v2) {
  return (vec4i_t){
      .x = v1.x * v2.x, .y = v1.y * v2.y, .z = v1.z * v2.z, .w = v1.w * v2.w};
}

vec4i_t vec_mul_num_vec4i_t(vec4i_t v1, i32 num) {
  return (vec4i_t){
      .x = v1.x * num, .y = v1.y * num, .z = v1.z * num, .w = v1.w * num};
}

vec4i_t vec_div_num_vec4i_t(vec4i_t v1, i32 num) {
  return (vec4i_t){
      .x = v1.x / num, .y = v1.y / num, .z = v1.z / num, .w = v1.w / num};
}

i32 vec_dot_vec4i_t(vec4i_t v1, vec4i_t v2) {
  return v1.x * v2.x + v1.y * v2.y + v1.z * v2.z + v1.w * v2.w;
}

i32 vec_magnitude_vec4i_t(vec4i_t v) {
  i32 dot_product = vec_dot_vec4i_t(v, v);
  return (i32)sqrtf((f32)dot_product);
}

vec4i_t vec_unit_vec4i_t(vec4i_t v) {
  i32 magnitude = vec_magnitude_vec4i_t(v);
  return (vec4i_t){v.x / magnitude, v.y / magnitude, v.z / magnitude,
                   v.w / magnitude};
}

vec4f_t vec_add_vec4f_t(vec4f_t v1, vec4f_t v2) {
  return (vec4f_t){
      .x = v1.x + v2.x, .y = v1.y + v2.y, .z = v1.z + v2.z, .w = v1.w + v2.w};
}

vec4f_t vec_sub_vec4f_t(vec4f_t v1, vec4f_t v2) {
  return (vec4f_t){
      .x = v1.x - v2.x, .y = v1.y - v2.y, .z = v1.z - v2.z, .w = v1.w - v2.w};
}

vec4f_t vec_mul_num_vec4f_t(vec4f_t v1, f32 num) {
  return (vec4f_t){
      .x = v1.x * num, .y = v1.y * num, .z = v1.z * num, .w = v1.w * num};
}

vec4f_t vec_div_num_vec4f_t(vec4f_t v1, f32 num) {
  return (vec4f_t){
      .x = v1.x / num, .y = v1.y / num, .z = v1.z / num, .w = v1.w / num};
}

vec4f_t vec_mul_vec4f_t(vec4f_t v1, vec4f_t v2) {
  return (vec4f_t){
      .x = v1.x * v2.x, .y = v1.y * v2.y, .z = v1.z * v2.z, .w = v1.w * v2.w};
}

f32 vec_dot_vec4f_t(vec4f_t v1, vec4f_t v2) {
  return v1.x * v2.x + v1.y * v2.y + v1.z * v2.z + v1.w * v2.w;
}

f32 vec_magnitude_vec4f_t(vec4f_t v) {
  f32 dot_product = vec_dot_vec4f_t(v, v);
  return sqrtf(dot_product);
}

vec4f_t vec_unit_vec4f_t(vec4f_t v) {
  f32 magnitude = vec_magnitude_vec4f_t(v);
  return (vec4f_t){v.x / magnitude, v.y / magnitude, v.z / magnitude,
                   v.w / magnitude};
}

mat4x4f_t get_scaling_matrix(vec3f_t scale) {
  // clang-format off
  return (mat4x4f_t){
      (vec4f_t){scale.x, 0.0f, 0.0f, 0.0f},
      (vec4f_t){0.0f, scale.y, 0.0f, 0.0f},
      (vec4f_t){0.0f, 0.0f, scale.z, 0.0f},
      (vec4f_t){0.0f, 0.0f,    0.0f, 1.0f},
  };
  // clang-format on
}

mat3x3f_t get_rotation_mat3x3f(vec3f_t rotation) {
  f32 x_rad = RADIANS(rotation.x);
  f32 y_rad = RADIANS(rotation.y);
  f32 z_rad = RADIANS(rotation.z);

  f32 cos_x = cosf(x_rad);
  f32 sin_x = sinf(x_rad);

  f32 cos_y = cosf(y_rad);
  f32 sin_y = sinf(y_rad);

  f32 cos_z = cosf(z_rad);
  f32 sin_z = sinf(z_rad);

  f32 row0_col0 = cos_z * cos_y;
  f32 row0_col1 = cos_z * sin_y * sin_x - sin_z * cos_x;
  f32 row0_col2 = cos_z * sin_y * cos_x + sin_z * sin_x;

  f32 row1_col0 = sin_z * cos_y;
  f32 row1_col1 = sin_z * sin_y * sin_x + cos_z * cos_x;
  f32 row1_col2 = sin_z * sin_y * cos_x - cos_z * sin_x;

  f32 row2_col0 = -sin_y;
  f32 row2_col1 = cos_y * sin_x;
  f32 row2_col2 = cos_y * cos_x;

  return (mat3x3f_t){
      .row0 = {row0_col0, row0_col1, row0_col2},
      .row1 = {row1_col0, row1_col1, row1_col2},
      .row2 = {row2_col0, row2_col1, row2_col2},
  };
}

mat4x4f_t get_translation_matrix(vec3f_t translation) {
  return (mat4x4f_t){
      (vec4f_t){1.0f, 0.0f, 0.0f, translation.x},
      (vec4f_t){0.0f, 1.0f, 0.0f, translation.y},
      (vec4f_t){0.0f, 0.0f, 1.0f, translation.z},
      (vec4f_t){0.0f, 0.0f, 0.0f, 1.0f},
  };
}

mat4x4f_t get_rotation_matrix(vec3f_t rotation) {
  mat3x3f_t mat = get_rotation_mat3x3f(rotation);

  // clang-format off
  return (mat4x4f_t){
      (vec4f_t){mat.row0.x, mat.row0.y, mat.row0.z, 0.0f},
      (vec4f_t){mat.row1.x, mat.row1.y, mat.row1.z, 0.0f},
      (vec4f_t){mat.row2.x, mat.row2.y, mat.row2.z, 0.0f},
      (vec4f_t){      0.0f,       0.0f,       0.0f, 1.0f},
  };
  // clang-format on
}

vec3f_t mul_mat3x4f_by_vec4f(mat3x4f_t mat, vec4f_t vec) {
  f32 x = mat.row0.x * vec.x + mat.row0.y * vec.y + mat.row0.z * vec.z +
          mat.row0.w * vec.w;
  f32 y = mat.row1.x * vec.x + mat.row1.y * vec.y + mat.row1.z * vec.z +
          mat.row1.w * vec.w;
  f32 z = mat.row2.x * vec.x + mat.row2.y * vec.y + mat.row2.z * vec.z +
          mat.row2.w * vec.w;

  return (vec3f_t){x, y, z};
}

mat3x4f_t mul_mat3x4f_by_mat4x4f(mat3x4f_t mat1, mat4x4f_t mat2) {
  f32 row0_col0 = mat1.row0.x * mat2.row0.x + mat1.row0.y * mat2.row1.x +
                  mat1.row0.z * mat2.row2.x + mat1.row0.w * mat2.row3.x;
  f32 row0_col1 = mat1.row0.x * mat2.row0.y + mat1.row0.y * mat2.row1.y +
                  mat1.row0.z * mat2.row2.y + mat1.row0.w * mat2.row3.y;
  f32 row0_col2 = mat1.row0.x * mat2.row0.z + mat1.row0.y * mat2.row1.z +
                  mat1.row0.z * mat2.row2.z + mat1.row0.w * mat2.row3.z;
  f32 row0_col3 = mat1.row0.x * mat2.row0.w + mat1.row0.y * mat2.row1.w +
                  mat1.row0.z * mat2.row2.w + mat1.row0.w * mat2.row3.w;

  f32 row1_col0 = mat1.row1.x * mat2.row0.x + mat1.row1.y * mat2.row1.x +
                  mat1.row1.z * mat2.row2.x + mat1.row1.w * mat2.row3.x;
  f32 row1_col1 = mat1.row1.x * mat2.row0.y + mat1.row1.y * mat2.row1.y +
                  mat1.row1.z * mat2.row2.y + mat1.row1.w * mat2.row3.y;
  f32 row1_col2 = mat1.row1.x * mat2.row0.z + mat1.row1.y * mat2.row1.z +
                  mat1.row1.z * mat2.row2.z + mat1.row1.w * mat2.row3.z;
  f32 row1_col3 = mat1.row1.x * mat2.row0.w + mat1.row1.y * mat2.row1.w +
                  mat1.row1.z * mat2.row2.w + mat1.row1.w * mat2.row3.w;

  f32 row2_col0 = mat1.row2.x * mat2.row0.x + mat1.row2.y * mat2.row1.x +
                  mat1.row2.z * mat2.row2.x + mat1.row2.w * mat2.row3.x;
  f32 row2_col1 = mat1.row2.x * mat2.row0.y + mat1.row2.y * mat2.row1.y +
                  mat1.row2.z * mat2.row2.y + mat1.row2.w * mat2.row3.y;
  f32 row2_col2 = mat1.row2.x * mat2.row0.z + mat1.row2.y * mat2.row1.z +
                  mat1.row2.z * mat2.row2.z + mat1.row2.w * mat2.row3.z;
  f32 row2_col3 = mat1.row2.x * mat2.row0.w + mat1.row2.y * mat2.row1.w +
                  mat1.row2.z * mat2.row2.w + mat1.row2.w * mat2.row3.w;

  return (mat3x4f_t){
      (vec4f_t){row0_col0, row0_col1, row0_col2, row0_col3},
      (vec4f_t){row1_col0, row1_col1, row1_col2, row1_col3},
      (vec4f_t){row2_col0, row2_col1, row2_col2, row2_col3},
  };
}

vec4f_t mul_mat4x4f_by_vec4f(mat4x4f_t mat, vec4f_t vec) {
  f32 x = mat.row0.x * vec.x + mat.row0.y * vec.y + mat.row0.z * vec.z +
          mat.row0.w * vec.w;
  f32 y = mat.row1.x * vec.x + mat.row1.y * vec.y + mat.row1.z * vec.z +
          mat.row1.w * vec.w;
  f32 z = mat.row2.x * vec.x + mat.row2.y * vec.y + mat.row2.z * vec.z +
          mat.row2.w * vec.w;
  f32 w = mat.row3.x * vec.x + mat.row3.y * vec.y + mat.row3.z * vec.z +
          mat.row3.w * vec.w;

  return (vec4f_t){x, y, z, w};
}

mat4x4f_t mul_mat4x4f(mat4x4f_t mat1, mat4x4f_t mat2) {
  f32 row0_col0 = mat1.row0.x * mat2.row0.x + mat1.row0.y * mat2.row1.x +
                  mat1.row0.z * mat2.row2.x + mat1.row0.w * mat2.row3.x;
  f32 row0_col1 = mat1.row0.x * mat2.row0.y + mat1.row0.y * mat2.row1.y +
                  mat1.row0.z * mat2.row2.y + mat1.row0.w * mat2.row3.y;
  f32 row0_col2 = mat1.row0.x * mat2.row0.z + mat1.row0.y * mat2.row1.z +
                  mat1.row0.z * mat2.row2.z + mat1.row0.w * mat2.row3.z;
  f32 row0_col3 = mat1.row0.x * mat2.row0.w + mat1.row0.y * mat2.row1.w +
                  mat1.row0.z * mat2.row2.w + mat1.row0.w * mat2.row3.w;

  f32 row1_col0 = mat1.row1.x * mat2.row0.x + mat1.row1.y * mat2.row1.x +
                  mat1.row1.z * mat2.row2.x + mat1.row1.w * mat2.row3.x;
  f32 row1_col1 = mat1.row1.x * mat2.row0.y + mat1.row1.y * mat2.row1.y +
                  mat1.row1.z * mat2.row2.y + mat1.row1.w * mat2.row3.y;
  f32 row1_col2 = mat1.row1.x * mat2.row0.z + mat1.row1.y * mat2.row1.z +
                  mat1.row1.z * mat2.row2.z + mat1.row1.w * mat2.row3.z;
  f32 row1_col3 = mat1.row1.x * mat2.row0.w + mat1.row1.y * mat2.row1.w +
                  mat1.row1.z * mat2.row2.w + mat1.row1.w * mat2.row3.w;

  f32 row2_col0 = mat1.row2.x * mat2.row0.x + mat1.row2.y * mat2.row1.x +
                  mat1.row2.z * mat2.row2.x + mat1.row2.w * mat2.row3.x;
  f32 row2_col1 = mat1.row2.x * mat2.row0.y + mat1.row2.y * mat2.row1.y +
                  mat1.row2.z * mat2.row2.y + mat1.row2.w * mat2.row3.y;
  f32 row2_col2 = mat1.row2.x * mat2.row0.z + mat1.row2.y * mat2.row1.z +
                  mat1.row2.z * mat2.row2.z + mat1.row2.w * mat2.row3.z;
  f32 row2_col3 = mat1.row2.x * mat2.row0.w + mat1.row2.y * mat2.row1.w +
                  mat1.row2.z * mat2.row2.w + mat1.row2.w * mat2.row3.w;

  f32 row3_col0 = mat1.row3.x * mat2.row0.x + mat1.row3.y * mat2.row1.x +
                  mat1.row3.z * mat2.row2.x + mat1.row3.w * mat2.row3.x;
  f32 row3_col1 = mat1.row3.x * mat2.row0.y + mat1.row3.y * mat2.row1.y +
                  mat1.row3.z * mat2.row2.y + mat1.row3.w * mat2.row3.y;
  f32 row3_col2 = mat1.row3.x * mat2.row0.z + mat1.row3.y * mat2.row1.z +
                  mat1.row3.z * mat2.row2.z + mat1.row3.w * mat2.row3.z;
  f32 row3_col3 = mat1.row3.x * mat2.row0.w + mat1.row3.y * mat2.row1.w +
                  mat1.row3.z * mat2.row2.w + mat1.row3.w * mat2.row3.w;

  return (mat4x4f_t){
      (vec4f_t){row0_col0, row0_col1, row0_col2, row0_col3},
      (vec4f_t){row1_col0, row1_col1, row1_col2, row1_col3},
      (vec4f_t){row2_col0, row2_col1, row2_col2, row2_col3},
      (vec4f_t){row3_col0, row3_col1, row3_col2, row3_col3},
  };
}
